proc Help { tip } {
  set w .help
  catch [destroy $w]
  toplevel $w -bd 0
  wm withdraw $w
  wm title $w "Help"

  pack [frame $w.cline -bd 1 -relief raised] -fill x -side top
  pack [Button $w.cline.exit -image iexit -command "destroy $w"] -side right 
  Balloon $w.cline.exit "Exit window"

  pack [labelframe $w.main -text " Help " -bd 2] -expand yes -fill both -padx 5 -pady 5
  pack [scrollbar $w.main.sby -command "$w.main.txt yview"] -side right -fill y
  pack [text $w.main.txt -width 63 -height 18 -insertontime 0 -bg #fafafa \
    -yscrollcommand "$w.main.sby set" -padx 10]
  bindtags $w.main.txt $w.main.txt
  
  bind $w.main.txt <1> {focus %W}
  bind $w.main.txt <Control-c> [list destroy $w]
  bind $w.main.txt <Down>  { %W yview scroll +1 units }
  bind $w.main.txt <Up>    { %W yview scroll -1 units }
  bind $w.main.txt <Prior> { %W yview scroll -1 page }
  bind $w.main.txt <Next>  { %W yview scroll +1 page }
  bind $w.main.txt <Home>  { %W mark set insert 1.0; %W see insert }
  bind $w.main.txt <End>   { %W mark set insert end; %W see insert }

  $w.main.txt tag configure <h1> -font [option get $w.main.txt bigFont $w.main.txt] \
                              -foreground blue4 -justify center
  $w.main.txt tag configure <h2> -foreground blue3 -justify center
  $w.main.txt tag configure <ul> -foreground blue3
  $w.main.txt tag configure <l1> -foreground magenta4
  $w.main.txt tag configure <u>  -underline yes

  set row 1
  set lnglast 0
  foreach l [HelpRead $w $tip] {
    set lline [HelpSyntax $row $l]
    set str   [lindex $lline 0]
    set ltags [lindex $lline 1]
    set lng   [string length $str]
    if { $lng > $lnglast } { set lnglast $lng }
    $w.main.txt insert end "$str\n"
    foreach { tg x1 x2 } $ltags {
      eval $w.main.txt tag add $tg $x1 $x2
    }
    incr row
  }
  
  if { $tip == "any" || $tip == "objects" } {
    foreach l [HelpRead $w xedj] {
      set lline [HelpSyntax $row $l]
      set str   [lindex $lline 0]
      set ltags [lindex $lline 1]
      set lng   [string length $str]
      if { $lng > $lnglast } { set lnglast $lng }
      $w.main.txt insert end "$str\n"
      foreach { tg x1 x2 } $ltags {
        eval $w.main.txt tag add $tg $x1 $x2
      }
      incr row
    }  
  }

  if { $row > 25 }     { set row 25 }
  if { $lnglast > 70 } { set lnglast 70 }
  $w.main.txt configure -width $lnglast -height $row  
  
  pack [frame $w.status -bd 1 -relief sunken] -side bottom -fill x
  pack [label $w.status.l -fg blue3 -anchor w] -fill x -side left
  WindowCenter $w
  wm deiconify $w
  focus $w.main.txt
}

proc HelpRead { w target } {
  global Main
  set lst {}
  Status $w "Read help ..."
  set fd [open $Main(src)/help.msg r]
  set flg 0
  while { ![eof $fd] } {
    set str [gets $fd]
    if { $flg && [regexp -- "^\\..*" $str ] } break
    if $flg {
      lappend lst $str
    }
    if [regexp -- "^\\.$target$" $str ] { set flg 1 }
  }    
  close $fd
  Status $w ""
  return $lst
}

proc HelpSyntax { row str} {
  set ls ""
  set ltags ""
  regsub -all -- "( )|(<\[^<>]*>)" $str " {&} " str
  foreach i $str {
    if [regexp -- "<.*>" $i] {
      if [IsBe tag$i] {
        unset tag$i
        lappend ltags $row.[lmath_sum [list_lengths $ls]]
      } else {
        set tag$i 1
        lappend ltags $i $row.[lmath_sum [list_lengths $ls]]
      }
    } else {
      lappend ls $i
    }
  }
  set lsr ""
  for { set i 0 } { $i < [llength $ls] } { incr i } {
    append lsr [lindex $ls $i]
  }
  return [list $lsr $ltags]
}
