proc LiteXedj { main w run } {
  global BindText
  LiteXedj:Init $w.txt
  pack [scrollbar $w.sby -bd 1 -command "$w.txt yview"] -side right -fill y
  pack [scrollbar $w.sbx -bd 1 -command "$w.txt xview" -orient hor] -side bottom -fill x
  pack [text $w.txt -bd 0 -xscrollcommand "$w.sbx set" -yscrollcommand "$w.sby set" \
          -wrap none -tabs 0 -bg #fafafa -maxundo 0 -undo 1] -fill both -expand yes
  LiteXedj:BindText $w.txt
  LiteXedj:InitTextTag $w.txt
  focus $w.txt
  bind $w.txt <1>         { 
    focus %W 
    %W mark set insert @%x,%y
  }
  set BindText($w.txt,run)  $run
  set BindText($w.txt,main) $main
}

proc LiteXedj:Init { w } {
  global SetKey Block Syntax
  #   
  array set SetKey {
    ^b  LiteXedj:FKCursorLeft
    ^f  LiteXedj:FKCursorRight
    ^p  LiteXedj:FKCursorUp
    ^n  LiteXedj:FKCursorDown
    ^z  LiteXedj:FKPreviousWord
    ^x  LiteXedj:FKNextWord

    ^u  LiteXedj:FKPreviousPage
    ^v  LiteXedj:FKNextPage
    ^a  LiteXedj:FKBeginLine
    ^e  LiteXedj:FKEndLine
    ^ku LiteXedj:FKBeginFile
    ^kv LiteXedj:FKEndFile

    ^d  LiteXedj:FKDeleteChar
    ^y  LiteXedj:FKDeleteLine
    ^w  LiteXedj:FKDeleteWordEnd
    ^o  LiteXedj:FKDeleteWordBegin
    ^j  LiteXedj:FKDeleteLineEnd
    ^-  LiteXedj:FKUndo
    ^6  LiteXedj:FKRedo

    ^2  LiteXedj:FKInsertSpace

    ^kb LiteXedj:FKBlockBegin
    ^kk LiteXedj:FKBlockEnd
    ^kc LiteXedj:FKBlockCopy
    ^ky LiteXedj:FKBlockDelete
    ^km LiteXedj:FKBlockMove
    ^k. LiteXedj:FKBlockRight
    ^k, LiteXedj:FKBlockLeft
    
    ^[c LiteXedj:FKRun
    ^c  LiteXedj:Exit
  }  

  array set Syntax {
    keyw     { create drop insert update on conflict primary not unique default \
               check references before after instead of for begin end replace \
               natural left right full union select when
             }
    opt      { table view trigger into values set where group by order limit index \
               unique asc desc temp temporary new old from index like glob null \
               offset having join outer inner on using all intersect except key as \
               each row
             }
    razd     { ] \[ \{ \} ( ) , . : ' ; -- }
    subst    { $ % \" & # }
    oper     { + - = == <= >= < > / != || ! && }
    datatype { integer smallint date char varchar serial decimal text boolean int \
               numeric character }
  }
  set Syntax($w,flgrun)       0
  set Block($w,flg,begin)     0
  set Block($w,data)          {}
}

#    Text
proc LiteXedj:BindText { w } {
  global BindText tcl_platform DoubleKey

  set Place($w,data)  ""
  set Place($w,poz)   -1

  set DoubleKey($w) " "

  bindtags $w [list $w .]

  bind $w <KeyPress> {
    set K %K

    #      
    if { "$K" == "Return" || "$K" == "space" || "$K" == "Up" || "$K" == "Down" } {
      set rowcol [%W index insert]
      scan $rowcol "%%\[^.]%%*c%%s" row col
      set str [%W get $row.0 {insert lineend}]
      LiteXedj:Syntax %W $row $str
    }
    
    if { "$K" == "Up"   || "$K" == "Down" || "$K" == "Left" || "$K" == "Right" || \
         "$K" == "Home" || "$K" == "End"  || "$K" == "Next" || "$K" == "Prior" } {
      LiteXedj:CursorMove $K %W 1
    }
    
    if { ( %s == 0 || %s == 8 || %s == 16 || %s == 1 || %s == 9 || %s == 17 ) && "%A" != "{}" } { 
      LiteXedj:InsertChar %W %A 
    }
    
    #    
    LiteXedj:PreEnterChars %W %s %K
    
    set rowcol [%W index insert]
    scan $rowcol "%%\[^.]%%*c%%s" row col    
    
    if { $tcl_platform(platform) == "windows" } {
      if { "$K" == "Delete" } { FKDeleteChar %W }
    }

    #   ^...
    if { %s == 4 || %s == 12 || %s == 20 } {
      if { "%K" == "minus" }       { set K "-" }
      if { "%K" == "equal" }       { set K "+" }
      if { "%K" == "period" }      { set K "." }
      if { "%K" == "comma" }       { set K "," }
      if { "%K" == "bracketleft" } { set K {\[} }
      if { $DoubleKey(%W) != " " } {
        #   ^xx
        regsub "\\\[" $DoubleKey(%W) "\\\\\[" DoubleKey(%W)
        set lsetkey [array get SetKey ^$DoubleKey(%W)$K]
        if { "$lsetkey" != "" } { eval [lindex $lsetkey 1] %W }
        set DoubleKey(%W) " "
      } else {
        set lsetkey [array get SetKey ^$K*]
        if { [llength $lsetkey] == 2 } {
          if { $DoubleKey(%W) == " " } {
            if { [string length [set firstelm [lindex $lsetkey 0]]] == 2 } {
              #   ^x
              eval [lindex $lsetkey 1] %W
            } else {
              set DoubleKey(%W) [string range $firstelm 1 1]
            }
          }
        } elseif { [llength $lsetkey] > 2 } {
          set DoubleKey(%W) [string range [lindex $lsetkey 0] 1 1]
        }
      }
    } else {
      set DoubleKey(%W) " "
    }
    
    #       
    if { "$K" == "Return" } {
      set prow [expr $row - 1]
      foreach i [split [LiteXedj:BeginSpacesInLine %W $prow] ""] {
        LiteXedj:InsertChar %W $i
      }
    }
  }

  bind $w <KeyRelease> {
    set TopLabel(%W,rowcol) [%W index insert]
    scan $TopLabel(%W,rowcol) "%%\[^.]%%*c%%s" row col
    
    #   
    if $Block(%W,flg,begin) {
      %W tag remove block 1.0 end
      %W tag add block $Block(%W,beginidx) insert
    }
    
    #   
    if { "%K" == "space" } {
      LiteXedj:PreEntry %W $row
    }
  }
}

proc LiteXedj:InitTextTag { w } {
  $w tag configure block      -background honeydew2
  $w tag configure razd       -foreground magenta3
  $w tag configure keyw       -foreground blue3
  $w tag configure subst      -foreground green4
  $w tag configure oper       -foreground orange3
  $w tag configure opt        -foreground blue4
  $w tag configure datatype   -foreground green4
  $w tag configure comment    -foreground grey60
}

proc LiteXedj:InsertChar { W A } {
  global Syntax
  
  set idx [$W index insert]
  set idxe $idx+1c
  if { "$A" == "\r" } {
    set A "\n"
  } elseif { "$A" == "\b" } {
    if [$W compare 1.0 != insert] {
      LiteXedj:DeleteChars $W $idx-1c
      set A ""
    }
  } elseif { "$A" == "\177" } {
    LiteXedj:DeleteChars $W $idx
    set A ""
  } elseif { "$A" == "\t" } {
    scan $idx "%\[^.]%*c%s" row col
    set A " "
    if ![expr $col % 2] {
      set A "  "
      set idxe $idx+2c
    }
  }
  if { "$A" != "" } {
    $W insert $idx $A
    $W tag remove find 1.0 end
  }
  if { "$A" == " "  || "$A" == "\n" || "$A" == "(" || "$A" == ")" || "$A" == "/" ||
       "$A" == "\[" || "$A" == "]"  || "$A" == "<" || "$A" == ">" || "$A" == ":" ||
       "$A" == "\"" || "$A" == "{"  || "$A" == "}" || "$A" == "." || "$A" == "_" } {
    $W edit separator
  }
  $W see insert
  set Syntax($W,flgrun) 1
}

proc LiteXedj:DeleteChars { W args } {
  global Syntax
  
  set idx0 [lindex $args 0]
  set idx1 [lindex $args 1]
  set idx0 [$W index $idx0]
  if { $idx1 != "" } {
    set idx1 [$W index $idx1]
  } else {
    set idx1 $idx0+1c
  }
  set A    [$W get $idx0 $idx1]
  if { "$A" == " "  || "$A" == "\n" || "$A" == "(" || "$A" == ")" || "$A" == "/" ||
       "$A" == "\[" || "$A" == "]"  || "$A" == "<" || "$A" == ">" || "$A" == ":" ||
       "$A" == "\"" || "$A" == "{"  || "$A" == "}" || "$A" == "." || "$A" == "_" } {
    $W edit separator
  }
  eval $W delete $args
  $W tag remove find 1.0 end
  set Syntax($W,flgrun) 1
}

#   -
proc LiteXedj:PreEnterChars { w s K } {
  if { $s == 0 || $s == 1 } {
    set idxch [$w index insert]
    set nch [$w get $idxch "$idxch+1c"]
    set pch [$w get "$idxch-2c" "$idxch-1c"]
    if { ($nch == ""   || $nch == " "  || $nch == "\n" || 
          $nch == "]"  || $nch == "\}" || $nch == ")"  ||
          $nch == ","  || $nch == "\"" ) &&
         ($pch == ""   || $pch == " "  || $pch == "\n" || 
          $pch == "\[" || $pch == "\{" || $pch == "("  ||
          $pch == ","  || $pch == "\"" )  } {
      if { "$K" == "bracketleft" } { set ch \] }
      if { "$K" == "braceleft" }   { set ch \} }
      if { "$K" == "quotedbl" }    { set ch \" }
      if { "$K" == "apostrophe" }  { set ch '  }
      if { "$K" == "less" }        { set ch >  }
    }
    if { "$K" == "parenleft" }   { set ch )  }
    if [info exists ch] {
      LiteXedj:InsertChar $w $ch
      $w mark set insert $idxch
    }
  }
}

#   () 
proc LiteXedj:PreEntry { w row } {
  set str [string trimright [$w get $row.0 {insert lineend}]]
  set prn {}
  set oldinsert [$w index insert]
  set nbsp [LiteXedj:BeginSpacesInLine $w $row]
  set nchar 0
  switch -regexp -- $str {
    "^ *create table$"    { 
      set prn "\n\(\n$nbsp  id       integer not null primary key,\n$nbsp  \n$nbsp\);\n\n" 
    }
    "^ *create view$"     { 
      set prn " () as \n$nbsp  select \n$nbsp    from \n$nbsp    where \n;\n\n" 
    }
    "^ *create trigger$"     { 
      set prn "\n$nbsp  before insert on \n$nbsp  for each row\n$nbsp    begin\n$nbsp    \n$nbsp    end\n;\n\n" 
    }
    "^ *create unique index$"     { 
      set prn "\n$nbsp  on  ()\n;\n\n" 
    }
    "^ *create index$"     { 
      set prn "\n$nbsp  on  ()\n;\n\n" 
    }
  }
  if { $prn != {} } {
    set idx0 [$w index insert]
    $w insert insert $prn
    set idx1 [$w index insert]
    $w mark set insert "$oldinsert + $nchar chars"
    for { set i $row } { $i < [expr $row + 8] } { incr i } {
      LiteXedj:SyntaxExec $w $i [$w get $i.0 "$i.0 lineend"]
    }
  }
}

proc LiteXedj:BeginSpacesInLine { w row } {
  scan [$w get $row.0 "$row.0 lineend"] "%\[ ]%s" pre fuf
  if ![info exists pre] { set pre {} }
  return $pre
}

proc LiteXedj:CursorMove { args } {
  global Syntax

  set poz   [lindex $args 0]
  set W     [lindex $args 1]
  set jum   [lindex $args 2]
  set bbox  [$W bbox insert]
  set dx    [lindex $bbox 0]
  set dy    [lindex $bbox 1]

  $W edit separator
  set Syntax($W,flgrun) 0

  switch -exact $poz {
    Left  { set idx insert-1c }
    Right { set idx insert+1c }
    Up    { set idx insert-${jum}l }
    Down  { set idx insert+${jum}l }
    Home  { set idx 0.0 }
    End   { set idx end }
    Next  { 
      $W yview scroll 1 pages
      set idx @$dx,$dy
    }
    Prior { 
      $W yview scroll -1 pages
      set idx @$dx,$dy
    }
    PrevWord { set idx "insert-1c wordstart" }
    NextWord { set idx "insert wordend" }
  }
  if { "$idx" == "@," } { set idx @2,2 }
  $W mark set insert $idx
  $W see insert
  return [$W index insert]
}

# CURSOR ################################################################################
proc LiteXedj:FKCursorLeft   { w } { LiteXedj:CursorMove Left $w }
proc LiteXedj:FKCursorRight  { w } { LiteXedj:CursorMove Right $w }
proc LiteXedj:FKCursorUp     { w } { LiteXedj:CursorMove Up $w 15 }
proc LiteXedj:FKCursorDown   { w } { LiteXedj:CursorMove Down $w 15 }
proc LiteXedj:FKPreviousWord { w } {
  set chn [$w get insert-1c insert]
  set idx [LiteXedj:CursorMove PrevWord $w]
  if { $chn != " " && $chn != "\n" }  { return 0 }
  set chp [$w get $idx-1c $idx]
  if {  $chp == " " || $chp == "\n" } { set chld [LiteXedj:FKPreviousWord $w] }
  if ![info exists chld]              { LiteXedj:CursorMove PrevWord $w }
  return 0
}
proc LiteXedj:FKNextWord { w } {
  set idx [LiteXedj:CursorMove NextWord $w]
  if [$w compare $idx == end-1c]    { return }
  set ch [$w get $idx-1c $idx]
  if {  $ch == " " || $ch == "\n" } { LiteXedj:FKNextWord $w }
}
proc LiteXedj:FKPreviousPage { w } { LiteXedj:CursorMove Prior $w }
proc LiteXedj:FKNextPage     { w } { LiteXedj:CursorMove Next $w }
proc LiteXedj:FKBeginLine    { w } { 
  $w mark set insert "insert linestart" 
  $w see insert
}
proc LiteXedj:FKEndLine      { w } { 
  $w mark set insert "insert lineend" 
  $w see insert
}
proc LiteXedj:FKBeginFile    { w } { LiteXedj:CursorMove Home $w }
proc LiteXedj:FKEndFile      { w } { LiteXedj:CursorMove End $w }

proc LiteXedj:FKDeleteChar   { w } { LiteXedj:DeleteChars $w insert }
proc LiteXedj:FKDeleteLine   { w } {
  set pop +1c
  set fend [$w index end]
  set lend [$w index "insert lineend"]
  if { [expr $fend - $lend] <= 1.0 } { set pop +0c }
  LiteXedj:DeleteChars $w "insert linestart" "$lend $pop"
  $w edit separator
}
proc LiteXedj:FKDeleteWordEnd { w args } {
  set ch [$w get insert insert+1c]
  if { $ch == " " || $ch == "\n" } {
    LiteXedj:DeleteChars $w insert
    if ![$w compare insert == end-1c] { LiteXedj:FKDeleteWordEnd $w chld }
  } elseif { $args != "chld" } {
    LiteXedj:DeleteChars $w insert "insert wordend"
  }
  $w edit separator
}
proc LiteXedj:FKDeleteWordBegin { w args } {
  set ch [$w get insert-1c insert]
  if { $ch == " " || $ch == "\n" } {
    LiteXedj:DeleteChars $w insert-1c
    if ![$w compare insert == 1.0] { LiteXedj:FKDeleteWordBegin $w chld }
  } elseif { $args != "chld" } {
    LiteXedj:DeleteChars $w "insert-1c wordstart" insert
  }
  $w edit separator
}
proc LiteXedj:FKDeleteLineEnd { w } {
  set ch [$w get insert insert+1c]
  if { $ch == "\n" } {
    LiteXedj:DeleteChars $w insert
  } else {
    LiteXedj:DeleteChars $w insert "insert lineend"
  }
  $w edit separator
}

proc LiteXedj:FKUndo { w } { catch [list $w edit undo] }
proc LiteXedj:FKRedo { w } { catch [list $w edit redo] }
proc LiteXedj:FKInsertSpace { w } {
  LiteXedj:InsertChar $w " "
  $w mark set insert insert-1c
}

proc LiteXedj:FKBlockBegin { w } {
  global Block
  $w tag remove block 1.0 end
  set Block($w,beginidx) [$w index insert]
  set Block($w,flg,begin) 1
}

proc LiteXedj:FKBlockEnd { w } {
  global Block
  set Block($w,flg,begin) 0
  set idxs [$w tag ranges block]
  if { $idxs != "" } { set Block($w,data) [eval $w get $idxs] }
}

proc LiteXedj:FKBlockCopy { args } {
  global Block
  set w       [lindex $args 0]
  set flgmove [lindex $args 1]

  if { $Block($w,data) != "" } {
    if { $flgmove == 1 } { 
      LiteXedj:FKBlockDelete $w
    }
    set idx0 [$w index insert]
    $w insert insert $Block($w,data)
    set idx1 [$w index insert]
    $w tag remove block 1.0 end
    $w tag add block $idx0 $idx1
    $w mark set insert $idx0
    $w see $idx0
    $w edit separator
    LiteXedj:SyntaxLines $w $idx0 $idx1    
  }
}

proc LiteXedj:FKBlockDelete { w } {
  if { [set idxs [$w tag ranges block]] != "" } {
    eval LiteXedj:DeleteChars $w $idxs
  }
}

proc LiteXedj:FKBlockMove { w } {
  LiteXedj:FKBlockCopy $w 1
}

proc LiteXedj:FKBlockRight { w } {
  set idxs [$w tag ranges block]  
  set idx0 [lindex $idxs 0]
  set idx1 [lindex $idxs 1]
  for { set i [file rootname $idx0] } { $i <= [file rootname $idx1] } { incr i } {
    $w insert $i.0 " "
  }
}

proc LiteXedj:FKBlockLeft { w } {
  set idxs [$w tag ranges block]  
  set idx0 [lindex $idxs 0]
  set idx1 [lindex $idxs 1]
  for { set i [file rootname $idx0] } { $i <= [file rootname $idx1] } { incr i } {
    LiteXedj:DeleteChars $w $i.0
  }
}

proc LiteXedj:SyntaxLines { w idxb idxe } {
  global Syntax
  scan $idxb "%\[^.]%*c%s" row col
  set strs [$w get "$idxb linestart" "$idxe lineend"]
  foreach { str } [split $strs "\n"] {
    LiteXedj:SyntaxExec $w $row $str
    incr row
  }
}

proc LiteXedj:Syntax { w row str } {
  global Syntax
  #      row
  if $Syntax($w,flgrun) {
    foreach tn [$w tag names] {
      if { "$tn" != "block" } {
        $w tag remove $tn $row.0 "$row.0 lineend"
      }
    }  
    LiteXedj:SyntaxExec $w $row $str
  }
}

proc LiteXedj:SyntaxExec { w row str } {
  global Syntax

  set Syntax(flgnameproc) 0
  set col 0
  if [regexp "(^ *--)|(^ *#)" $str] {
    $w tag add comment $row.0 "$row.0 lineend"
  } else {
    foreach word [split $str { }] {
      if { !$Syntax(flgnameproc) && [string length $word] > 1 && 
           ![regexp "^-.+" $word] &&
           [regexp "\{|\}|\\\[|]|<|>|\\$|\"|\\(|\\)|%|\\.|-|:|'|," $word] } {
        foreach word [LiteXedj:SplitWord $word] {
          set colend [expr $col + [string length $word]]
          LiteXedj:SyntaxSet $w $word $row $col $colend
          set col $colend
        }
      } else {
        set colend [expr $col + [string length $word]]
        LiteXedj:SyntaxSet $w $word $row $col $colend
        set col $colend
      }
      incr col
    }
  }
}

proc LiteXedj:SplitWord { mword } {
  set word {}
  foreach ch [split $mword {}] {
    if [string is wordchar $ch] {
      append word $ch
    } else {
      append word " [list $ch] "
    }
  }
  return $word
}

proc LiteXedj:SyntaxSet { w word row col colend } {
  global Syntax
  regsub -all "\\\[" $word "\\\\\[" word
  if { [lsearch $Syntax(opt) $word] > -1 } {
    $w tag add opt $row.$col $row.$colend
  } elseif { [lsearch $Syntax(razd) $word] > -1 } {
    $w tag add razd $row.$col $row.$colend
  } elseif { [lsearch $Syntax(keyw) $word] > -1 } {
    $w tag add keyw $row.$col $row.$colend
  } elseif { [lsearch $Syntax(subst) $word] > -1 } {
    $w tag add subst $row.$col $row.$colend
  } elseif { [lsearch $Syntax(oper) $word] > -1 } {
    $w tag add oper $row.$col $row.$colend
  } elseif { [lsearch $Syntax(datatype) $word] > -1 } {
    $w tag add datatype $row.$col $row.$colend
  }
}

proc LiteXedj:FKRun { w } {
  global BindText
  eval $BindText($w,run)
}

proc LiteXedj:Exit { w } {
  global BindText
  destroy $BindText($w,main)
}
