proc ObjectsAll { } {
  set w .objects
  catch [destroy $w]
  toplevel $w -bd 0
  wm withdraw $w
  wm title $w "Objects DB"

  pack [frame $w.cline -bd 1 -relief raised] -fill x -side top
  pack [frame $w.nb -bd 0 -relief raised] -fill x -side top -pady 2
  pack [frame $w.status -bd 1 -relief sunken] -side bottom -fill x

  pack [Button $w.cline.data -image idata] -side left
  Balloon $w.cline.data "Table data"
  pack [Button $w.cline.drop -image idrop] -side left
  Balloon $w.cline.drop "Object drop"
  pack [Separator $w.cline.sep1] -side left -fill y -padx 3

  pack [Button $w.cline.execute -image iexecute] -side left
  Balloon $w.cline.execute "Execute query"
  pack [Button $w.cline.exit -image iexit -command "destroy $w"] -side right 
  Balloon $w.cline.exit "Exit window"
  pack [Separator $w.cline.sepend] -side right -fill y -padx 3

  pack [Button $w.cline.help -image ihelp -command "Help objects"] -side right
  Balloon $w.cline.help "Help"

  #  NoteBook
  
  pack [frame $w.nb.startLine -bg white -width 7] -fill x -anchor s -side left
  foreach i { Table Index View Trigger } {
    pack [nbButton $w.nb.w_$i -text $i] -side left -anchor s
  }
  $w.nb.w_Table subwdg FrameB "configure -bg grey85"
  $w.nb.w_Table subwdg FrameR "configure -bg grey50"
  $w.nb.w_Table subwdg Button "configure -pady 3"
  pack [frame $w.nb.endLine -bg white] -fill x -anchor s -side bottom

  Objects:InitMenu $w
  Objects Table
  
  pack [label $w.status.l -fg blue3 -anchor w] -fill x -side left
  WindowCenter $w
  wm deiconify $w
}

proc Objects { tip } {
  set w .objects.w_$tip

  set pwpack "-fill both -expand yes"
  .objects.cline.execute subwdg Button "configure -command \"Objects:RightExecute $w $w.rf.txt $w.lf.tr $tip\""
  
  if [winfo exists $w] {
    Objects:ADcline $w.lf.tr $tip
    eval pack $w $pwpack
  } else {
    eval pack [panedwindow $w -sashcursor sb_h_double_arrow] $pwpack
    labelframe $w.lf -text " Object $tip " -bd 2
    labelframe $w.rf -text " Source " -bd 2
    $w add $w.lf -padx 5 -pady 5 -minsize 100 -width 200
    $w add $w.rf -padx 5 -pady 5 -minsize 100 -width 500
    
    # left window
    pack [TreeWdg $w.lf.tr -width 40 -height 20 -bd 1 \
      -command "Objects:Tree:SelChailds $w.lf.tr $tip"] -fill both -expand yes
    $w.lf.tr bind <Visibility> { focus %W }
    
    # right window
    LiteXedj .objects $w.rf "Objects:RightExecute $w $w.rf.txt $w.lf.tr $tip"
    
    $w.lf.tr bind <1>              [list Objects:TreeWdg:B1 $w.lf.tr $w.rf $tip]
    $w.lf.tr bind <Control-Return> [list Objects:TreeWdg:B1 $w.lf.tr $w.rf $tip]
    $w.lf.tr bind <Control-c>      { destroy .objects }
    $w.lf.tr bind <3>              [list Objects:TreeWdg:B3 $w.lf.tr $w.rf $tip %W %x %y %X %Y]
    
    Objects:LeftRead $w.lf.tr $tip
    Objects:ADcline  $w.lf.tr $tip
  }
}

proc Objects:ADcline { w tip } {
  set node    [$w node]
  set parent  [$w parent $node]
  set pparent [$w parent $parent]
  foreach i { data drop execute } {
    .objects.cline.$i subwdg Button "configure -state normal"
  }
  if { $tip == "Table" } {
    if { "$parent" == "root" } {
      .objects.cline.data subwdg Button "configure -state normal -command \"Table:Data $node\""
      .objects.cline.drop subwdg Button "configure -command \"Objects:Delete $node $w Table Table\""
    } else {
      .objects.cline.data subwdg Button "configure -state disable"
      if { "$parent" == "$pparent:index" } {
        .objects.cline.drop subwdg Button "configure -command \"Objects:Delete $node $w Table Index\""
      }
      if { "$parent" == "$pparent:trigger" } {
        .objects.cline.drop subwdg Button "configure -command \"Objects:Delete $node $w Table Trigger\""
      }
    }
  } else {
    .objects.cline.data subwdg Button "configure -state disable"
    .objects.cline.drop subwdg Button "configure -command \"Objects:Delete $node $w $tip $tip\""
  }
  if { "$node" == "$parent:index" || "$node" == "$parent:trigger" } {
    foreach i { data drop execute } {
      .objects.cline.$i subwdg Button "configure -state disable"
    }
  }
}

proc Objects:InitMenu { w } {
  menu $w.menuTable -tearoff 0 -disabledforeground white
    $w.menuTable add command -label "Table operations" -state disable -background #9090f0
    $w.menuTable add command -label "table data"
    $w.menuTable add command -label "table drop"
    $w.menuTable add command -label "table source"
    $w.menuTable add separator
    $w.menuTable add command -label exit
  
  foreach i { Index View Trigger } {
    menu $w.menu$i -tearoff 0 -disabledforeground white
      $w.menu$i add command -label "$i operations" -state disable -background #9090f0
      $w.menu$i add command -label "[string tolower $i] drop"
      $w.menu$i add command -label "[string tolower $i] source"
      $w.menu$i add separator
      $w.menu$i add command -label exit
  }
}

proc Objects:TreeWdg:B3 { wtr wrf tip W x y X Y } {
  Tree::ActivateLine $W "[$W index @$x,$y] linestart" 
  set node    [$wtr node]
  set parent  [$wtr parent $node]
  set pparent [$wtr parent $parent]
  Objects:ADcline $wtr $tip
  if { $tip == "Table" } {
    if { $parent == "root" }             { 
      .objects.menuTable entryconfigure "table data"   -command "Table:Data $node"
      .objects.menuTable entryconfigure "table drop" -command "Objects:Delete $node $wtr Table Table"
      .objects.menuTable entryconfigure "table source" -command "Objects:TreeWdg:B1 $wtr $wrf $tip"
      .objects.menuTable post $X $Y 
    }
    if { $parent == "$pparent:index" }   { 
      .objects.menuIndex entryconfigure "index drop" -command "Objects:Delete $node $wtr Table Index"
      .objects.menuIndex entryconfigure "index source" -command "Objects:TreeWdg:B1 $wtr $wrf $tip"
      .objects.menuIndex   post $X $Y 
    }
    if { $parent == "$pparent:trigger" } { 
      .objects.menuTrigger entryconfigure "trigger drop" -command "Objects:Delete $node $wtr Table Trigger"
      .objects.menuTrigger entryconfigure "trigger source" -command "Objects:TreeWdg:B1 $wtr $wrf $tip"
      .objects.menuTrigger post $X $Y 
    }
  } else {
    .objects.menu$tip entryconfigure "[string tolower $tip] drop" -command "Objects:Delete $node $wtr $tip $tip"
    .objects.menu$tip entryconfigure "[string tolower $tip] source" -command "Objects:TreeWdg:B1 $wtr $wrf $tip"
    .objects.menu$tip post $X $Y
  }
  
}

proc Objects:Delete { node wtr tip subtip } {
  if { $tip == "Table" } {
    if { $subtip == "Table" } {
      set st [TableCmd:Delete .objects $node]
    } else {
      set st [Objects:Drop .objects $subtip $node]
    }
    if { $st == 1 } { Objects:LeftRead $wtr $tip }
  } else {
    set st [Objects:Drop .objects $tip $node]
    if { $st == 1 } { Objects:LeftRead $wtr $tip }
  }
}

proc Objects:Drop { w objtype objname } {
  set status {}
  if [tk_dialog $w.dlg question "Drop $objtype $objname ?" question 0 No Yes] {
    Status $w "Drop $objtype $objname ..."
    set err [catch "dbs eval \"drop $objtype $objname\"" error]
    if { $err } {
      tk_dialog $w.dlg error $error error 0 Ok
      set status -1
    } else {
      set status 1
    }
    Status $w ""
  }
  return $status
}

proc Objects:TreeWdg:B1 { w xedj tip } {
  Status .objects "Read source SQL ..."
  set node     [$w node]
  set parent   [$w parent $node]
  set nodename [$w cget $node -text]
  set sql      [$w cget $node -data]
  $xedj.txt delete 0.0 end
  $xedj.txt insert end $sql
  set rowe [lindex [split [$xedj.txt index end] "."] 0]
  for { set row 1 } { $row < $rowe } { incr row } {
    LiteXedj:SyntaxExec $xedj.txt $row [$xedj.txt get $row.0 "$row.0 lineend"]
  }

  Objects:ADcline $w $tip
  Status .objects ""
}

proc Objects:RightExecute { main w wtr tip } {
  TableCmd:Any:Execute $main $w
  Objects:LeftRead $wtr $tip
}

proc Objects:LeftRead { w tip } {
  #   , ,   
  set tp [string tolower $tip]

  if { $tip == "Table" } {
    set lsel [dbs eval "select distinct a.name, a.sql, b.tbl_name \
                          from sqlite_master as a left outer join sqlite_master as b 
                                  on a.name = b.tbl_name and a.type != b.type \
                          where a.type='$tp' \
                          order by a.name "] 
  } else {
    set lsel [dbs eval "select distinct a.name, a.sql, '' \
                          from sqlite_master as a
                          where a.type='$tp' \
                          order by a.name "]
  }
  
  set folderopened [$w opened]
  set aidx [Tree::LSearchNP [$w.txt dump -tag 1.0 end] active +1]
  $w delete root
  foreach { name sql tbl_name } $lsel {
    $w insert root $name -text $name -data $sql
    if { "$tbl_name" == "" } {
      $w configure $name -image sheet
    }
  }
  foreach node $folderopened {
    $w open $node
  }
  #     
  if { "$aidx" != "" } {
    Tree::ActivateLine $w.txt [lindex $aidx 0]
  }
  Objects:ADcline $w $tip
  #    
  [winfo parent [winfo parent $w]].rf.txt delete 0.0 end
}

proc Objects:Tree:SelChailds { node data args } {
  set w    [lindex $args 0]
  set tip  [lindex $args 1]

  set tp [string tolower $tip]
  if { [$w cget $node -image] == "sheet" } { 
    return 0 
  } else {
    # for table
    if { $tip == "Table" } {
      if { [$w parent $node] == "root" } {
        if { [$w chailds $node] == "" } {
          set lsel [dbs eval "select distinct type, '', '', '' from sqlite_master \
                                where tbl_name ='$node' and type != '$tp' \
                                order by type "]
          #    (index, trigger, view)   - $node
          #  
          for { set i 0 } { $i < [llength $lsel] } { incr i 4} {
            set ni [expr $i+1]
            set nd [lindex $lsel $i]
            set lsel [lreplace $lsel $i $ni $node:$nd $nd]
          }
          return $lsel 
        } else {
          return 1
        }
      } else {
        if { [$w chailds $node] == "" } {
          set nd [lindex [split $node ":"] 1]
          set lsel [dbs eval "select name, name, sql, 'sheet' from sqlite_master \
                                where type ='$nd' and tbl_name == '[$w parent $node]' \
                                order by name "]
          return $lsel
        } else {
          return 1
        }
      }
    }
  }
}
