proc Table:Data { tbl } {
  global Table ${tbl}_DATA $tbl

  catch "unset ${tbl}_DATA"
  set w .table_$tbl
  set cols [llength $Table($tbl,fields)]
  set width 7
  if { $cols < $width } { set width $cols }

  catch [destroy $w]
  toplevel $w -bd 0
  wm withdraw $w
  wm resizable $w 1 0
  wm title $w $tbl
  pack [frame $w.top -bd 1 -relief raised] -fill x -side top
  pack [frame $w.cline -bd 1 -relief raised] -fill x -side top
  pack [frame $w.status -bd 1 -relief sunken] -side bottom -fill x
  pack [labelframe $w.main -text " Table data " -bd 2] -expand yes -fill x -padx 5 -pady 5
  
  pack [LabelEntry $w.top.tbl -label Table: -highlightthickness 0] -anchor w
  $w.top.tbl subwdg Entry "insert end $tbl"
  $w.top.tbl subwdg Entry "configure -bd 0 -fg red3 -state readonly -width [string length $tbl]"
  
  pack [Button $w.cline.delete  -image idelete  -command "Table:Delete $w.main.tbl $tbl $cols"] -side left
  Balloon $w.cline.delete "Delete current row"
  pack [Button $w.cline.exit    -image iexit    -command "destroy $w"] -side right 
  Balloon $w.cline.exit "Exit window"
  pack [Separator $w.cline.sepend] -side right -fill y -padx 3
  pack [Button $w.cline.help    -image ihelp    -command "Help tabledata"] -side right
  Balloon $w.cline.help "Help"
  
  pack [label $w.status.l -fg blue3 -anchor w] -fill x -side left
  pack [scrollbar $w.main.sby -command "$w.main.tbl yview"] -side right -fill y
  pack [scrollbar $w.main.sbx -command "$w.main.tbl xview" -orient hor] -side bottom -fill x
  pack [table $w.main.tbl -width $width -height 21 -cols $cols -rows 21 -bd 0 \
      -colstretchmode all \
      -yscrollcommand "$w.main.sby set" \
      -xscrollcommand "$w.main.sbx set" \
      -browsecommand  "Table:Browsecommand $tbl %W $cols" \
      -rowtagcommand { TkTableRowTagCmd } \
      -variable ${tbl}_DATA \
      -titlerows 1 \
      -roworigin -1 \
      -colorigin 0] -side left -expand yes -fill x

  TkTableInitTag  $w.main.tbl
  TkTableInitBind $w.main.tbl

  bind $w <Control-c>          "$w.cline.exit subwdg Button invoke"
  bind $w.main.tbl <Control-d> "$w.cline.delete subwdg Button invoke"
  
  Table:Select $w.main.tbl $tbl $cols
  focus $w.main.tbl
  WindowCenter $w
  wm deiconify $w
}

proc Table:Browsecommand { tbl w cols } {
  global $tbl ${tbl}_DATA
  set active [$w index active]
  if ![IsBe ${tbl}(oldactive)] { set ${tbl}(oldactive) $active }
  set oldactive [subst $${tbl}(oldactive)]
  set fuf [split $active ","];      set row    [lindex $fuf 0];   set col    [lindex $fuf 1]
  set fuf [split $oldactive ","];   set oldrow [lindex $fuf 0];   set oldcol [lindex $fuf 1]

  # after row
  if { $oldrow != $row } {
    Table:AfterRow $tbl $w $oldrow $cols
  }
  set ${tbl}(oldactive) $active
}

proc Table:Select { w tbl cols } {
  global ${tbl}_DATA $tbl ${tbl}_OLDDATA Table

  Status .table_$tbl "Read data ... "
  array unset ${tbl}_DATA
  array unset ${tbl}_OLDDATA
  array unset $tbl
  set c 0
  #    
  foreach i $Table($tbl,fields) {
    set ${tbl}_DATA(-1,$c) $i
    set ${tbl}(lng,$c) [string length $i]
    $w tag col AnchorW $c
    incr c
  }
  for { set i 0 } { $i < $cols } { incr i } {
    append fields " f$i"
  }

  #    
  set lsel [dbs eval "select * from $tbl order by 1"]
  set row 0
  foreach $fields $lsel {
    for { set i 0 } { $i < $cols } { incr i } {
      set vr f$i
      set ${tbl}_DATA($row,$i) [subst $$vr]
      set slng [string length [subst $$vr]]
      if { [subst $${tbl}(lng,$i)] < $slng } { set ${tbl}(lng,$i) $slng }
    }
    incr row
  }

  #     
  if { $row > [expr [$w cget -rows] - 1] } {
    $w configure -rows [expr $row + 1]
  } else {
    $w configure -rows [$w cget -height]
  }
  #   
  for { set i 0 } { $i < $cols } { incr i } {
    $w width $i [subst $${tbl}(lng,$i)]
  }
  Status .table_$tbl ""
  $w activate 0,0
  array set ${tbl}_OLDDATA [array get ${tbl}_DATA]
}

proc Table:AfterRow { tbl w row cols } {
  global ${tbl}_DATA $tbl ${tbl}_OLDDATA
  set flgok 0
  for { set i 0 } { $i < [expr $cols - 1]} { incr i } {
    append ifins "![IsBe ${tbl}_OLDDATA($row,$i)] && "
  }
  append ifins "![IsBe ${tbl}_OLDDATA($row,$i)] && ("

  for { set i 0 } { $i < [expr $cols - 1]} { incr i } {
    append ifins "[IsBe ${tbl}_DATA($row,$i)] || "
    if ![IsBe ${tbl}_DATA($row,$i)]    { set ${tbl}_DATA($row,$i) "" }
    if ![IsBe ${tbl}_OLDDATA($row,$i)] { set ${tbl}_OLDDATA($row,$i) "" }
    append ifupd "\"[subst $${tbl}_DATA($row,$i)]\" != \"[subst $${tbl}_OLDDATA($row,$i)]\" || "
  }
  append ifins "[IsBe ${tbl}_DATA($row,$i)])"
  if ![IsBe ${tbl}_DATA($row,$i)]    { set ${tbl}_DATA($row,$i) "" }
  if ![IsBe ${tbl}_OLDDATA($row,$i)] { set ${tbl}_OLDDATA($row,$i) "" }
  append ifupd "\"[subst $${tbl}_DATA($row,$i)]\" != \"[subst $${tbl}_OLDDATA($row,$i)]\""

  if $ifins {
    set flgok [Table:Insert $tbl $row $cols]
  } elseif $ifupd {
    set flgok [Table:Update $tbl $row $cols]
  }

  if $flgok {
    for { set i 0 } { $i < $cols} { incr i } {
      if [IsBe ${tbl}_DATA($row,$i)] {
        set ${tbl}_OLDDATA($row,$i) [subst $${tbl}_DATA($row,$i)]
      } else {
        unset -nocomplain ${tbl}_OLDDATA($row,$i)
      }
    }  
  }
}

proc Table:Insert { tbl row cols } {
  global ${tbl}_DATA
  Status .table_$tbl "Insert data ..."
  for { set i 0 } { $i < [expr $cols -1] } { incr i } {
    append val "'[subst $${tbl}_DATA($row,$i)]',"
  }
  append val "'[subst $${tbl}_DATA($row,$i)]'"
  set err [catch "dbs eval \"insert into $tbl values($val)\"" error]
  Status .table_$tbl ""
  if { $err } {
    tk_dialog .table_$tbl.dlg error $error error 0 Ok
    return 0
  }
  return 1
}

proc Table:Update { tbl row cols } {
  global ${tbl}_DATA Table ${tbl}_OLDDATA
  Status .table_$tbl "Update data ..."
  set i 0
  foreach { fname } $Table($tbl,fields) {
    append val "$fname = '[subst $${tbl}_DATA($row,$i)]',"
    incr i
  }
  set val [string range $val 0 end-1]
  set err [catch "dbs eval \"update $tbl set $val \
            where [lindex $Table($tbl,fields) 0] = '[subst $${tbl}_OLDDATA($row,0)]'\"" error]
  Status .table_$tbl ""
  if { $err } {
    tk_dialog .table_$tbl.dlg error $error error 0 Ok
    return 0
  }
  return 1
}

proc Table:Delete { w tbl cols } {
  global ${tbl}_OLDDATA Table
  set active [$w index active]
  set row [lindex [split $active ","] 0]
  if [IsBe ${tbl}_OLDDATA($row,0)] {
    if [tk_dialog $w.dlg question "Delete active row ?" question 0 No Yes] {
      Status .table_$tbl "Delete data ..."
      dbs eval "delete from $tbl where [lindex $Table($tbl,fields) 0] = '[subst $${tbl}_OLDDATA($row,0)]'"
      Table:Select $w $tbl $cols
      $w activate $active
      Status .table_$tbl ""
    }
  }
}
